import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx
import connectors_core.alteryx.helpers as alteryx_helpers

from connectors_core.bigquery import BigQuerySchemaReader, SchemaConversionException

from connectors_core.bigquery.base import AbstractBigQueryPlugin
from connectors_core.bigquery.common import (
                     DEFAULT_MAX_FIELD_LENGTH,
                     BIG_QUERY_API_BASE)

from google.cloud import bigquery
from google.api_core.exceptions import BadRequest, NotFound, Forbidden, GoogleAPIError
from google.auth.exceptions import TransportError, RefreshError, DefaultCredentialsError

from typing import Union


class BigQueryInputEngine(AbstractBigQueryPlugin):
    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.auth_type = None   # type: str
        self.service_file_loc = None # type: str
        self.service_account_json = None # type: str
        self.project_id = None   # type: str
        self.dataset_id = None # type: str
        self.table_id = None # type: str
        self.credentials = None # type: google.auth.credentials.Credentials
        self.record_limit = 0
        self.should_limit_records = False
        self.max_field_length = DEFAULT_MAX_FIELD_LENGTH
        self.using_custom_query = False
        self.custom_query = None
        self.sql_type = None
        self.pre_sql_query = None
        self.post_sql_query = None

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        self.display_error_msg(
            self.alteryx_engine.xmsg("This tool does not accept incoming connections")
        )
        return None

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)
        self.auth_type = self.get_xml_config_value(str_xml, 'authType')
        if self.auth_type == 'EndUser':
            if not self.init_end_user_credentials(str_xml):
                return
        elif self.auth_type == 'ServicePrincipal':
            if not self.init_service_principal_credentials(str_xml):
                return
        else:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid authentication mode selected: @1', self.auth_type))
            return

        selected_id = self.get_xml_config_value(str_xml, 'selectedId')
        self.using_custom_query = self.get_xml_config_value(str_xml, 'usingCustomFilter') == 'True'
        if not self.using_custom_query:
            try:
                self.project_id, remainder = selected_id.split(':')
                self.dataset_id, self.table_id = remainder.split('.')
            except (AttributeError, ValueError) as e:
                self.invalidate_config(self.alteryx_engine.xmsg('Invalid table selection: please select a table.'))
        else:
            self.custom_query = self.get_xml_config_value(str_xml, 'inputCustomFilter')
            if not self.custom_query:
                self.invalidate_config(self.alteryx_engine.xmsg('Please enter a non-empty custom query.'))
            self.project_id = self.get_xml_config_value(str_xml, 'customFilterProjectId')
            if not self.project_id:
                self.invalidate_config(self.alteryx_engine.xmsg('Please select a project to run your custom query against.'))

        self.sql_type = self.get_xml_config_value(str_xml, 'sqlType')
        self.pre_sql_query = self.get_xml_config_value(str_xml, 'preSqlQuery')
        if self.pre_sql_query == 'Empty':
            self.pre_sql_query = None
        self.post_sql_query = self.get_xml_config_value(str_xml, 'postSqlQuery')
        if self.post_sql_query == 'Empty':
            self.post_sql_query = None

        try:
            self.max_field_length = int(self.get_xml_config_value(str_xml, 'maxFieldLength'))
        except (ValueError, TypeError) as e:
            self.max_field_length = DEFAULT_MAX_FIELD_LENGTH

        self.should_limit_records = self.get_xml_config_value(str_xml, 'shouldLimitRecords')
        if self.should_limit_records == 'True':
            try:
                self.record_limit = int(self.get_xml_config_value(str_xml, 'inputRecordLimit'))
            except TypeError:
                self.record_limit = -1
        else:
            self.record_limit = -1

        metadata_dict = {
            'Host': BIG_QUERY_API_BASE,
            'ServiceName': 'BigQuery',
            'PluginName': 'BigQuery',
        }
        if self.using_custom_query:
            metadata_dict['Query'] = self.custom_query
        if not self.custom_query:
            metadata_dict['Database'] = self.dataset_id
            metadata_dict['Table'] = self.table_id
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        return True

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        if not self.is_initialized:
            return False

        update_only = self.alteryx_engine.get_init_var(self.n_tool_id, 'UpdateOnly') == 'True' \
            or n_record_limit == 0
        record_limit = self.record_limit if self.record_limit > 0 else n_record_limit
        if record_limit == 0:
            update_only = True

        client = bigquery.Client(project=self.project_id, credentials=self.credentials)
        try:
            if not update_only and self.pre_sql_query:
                self.display_info_msg("Running pre-sql query")
                job_config = bigquery.QueryJobConfig()
                job_config.use_legacy_sql = self.sql_type == 'LEGACY_SQL'
                # Run pre sql query
                query_job = client.query(self.pre_sql_query, job_config=job_config)
                query_job.result()
                self.display_info_msg("Pre-sql query successfully finished")

            schema_reader, row_iterator = None, None
            if not self.using_custom_query:
                schema_reader, row_iterator, total_records = self._push_records_from_table(client, update_only)
            else:
                schema_reader, row_iterator, total_records = self._push_records_from_query(client, update_only)

            record_info = schema_reader.get_record_info()
            self.output_anchor.init(record_info)

            if update_only:
                return True

            record_creator = record_info.construct_record_creator()  # type: Sdk.RecordCreator
            record_count = 0
            zero_total_records_warning_shown = False if not self.using_custom_query else True
            for result in row_iterator:
                if total_records == 0 and not zero_total_records_warning_shown:
                    self.display_warning_msg(
                        self.alteryx_engine.xmsg(
                            'Obtained unknown number of records from BigQuery. '
                            'Unable to estimate progress indicator while workflow is running.'
                        )
                    )
                    zero_total_records_warning_shown = True
                record_info_out = record_info.clone()  # type: Sdk.RecordInfo
                record_creator.reset()
                for field in row_iterator.schema:
                    for record_field, field_value in schema_reader.generate_records_from_field(result, field, record_info_out, ""):
                        alteryx_helpers.set_value_on_field(record_field, record_creator, field_value, self.alteryx_engine)

                finalized_record = record_creator.finalize_record()
                self.output_anchor.push_record(finalized_record)
                record_count += 1
                if total_records != 0:
                    self.output_anchor.update_progress(record_count / total_records)

            if not update_only and self.post_sql_query:
                self.display_info_msg("Running post-sql query")
                job_config = bigquery.QueryJobConfig()
                job_config.use_legacy_sql = self.sql_type == 'LEGACY_SQL'
                # Run post sql query
                query_job = client.query(self.post_sql_query, job_config=job_config)
                query_job.result()
                self.display_info_msg("Post-sql query successfully finished")

            finish_msg = self.alteryx_engine.xmsg(
                "@1 @2 records were read from @3:@4", str(record_count), self.table_id, self.project_id, self.dataset_id
            )
            if self.using_custom_query:
                finish_msg = self.alteryx_engine.xmsg(
                    "@1 records were read from a custom BigQuery Query.", str(record_count)
                )
            self.display_info_msg(finish_msg)

            self.output_anchor.update_progress(1.0)
            self.output_anchor.close()  # Close outgoing connections.
            return True
        except BadRequest as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
        except NotFound as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Bad Request to Google API: @2", str(e.code.value), e.message)
            )
        except Forbidden as e:
            self.log.exception(e)
            self.display_error_msg(
                self.alteryx_engine.xmsg("@1 Invalid Permissions: @2", str(e.code.value), e.message)
            )
        except GoogleAPIError as e:
            self.log.exception(e)
            self.display_error_msg(self.alteryx_engine.xmsg("GoogleAPIError received: @1", str(e)))
        except TransportError as e:
            self.log.exception(
                self.alteryx_engine.xmsg("TransportError received. Check your proxy credentials for accuracy.")
            )
            self.display_error_msg(
                self.alteryx_engine.xmsg("TransportError received. Check your proxy credentials for accuracy.")
            )
        except RefreshError as e:
            self.log.exception(
                self.alteryx_engine.xmsg("RefreshError received. Check your credentials for accuracy.")
            )
            self.display_error_msg(
                self.alteryx_engine.xmsg("RefreshError received. Check your credentials for accuracy.")
            )
        except DefaultCredentialsError as e:
            self.log.exception(
                self.alteryx_engine.xmsg("DefaultCredentialsError received. Check your credentials for accuracy.")
            )
            self.display_error_msg(
                self.alteryx_engine.xmsg("DefaultCredentialsError received. Check your credentials for accuracy.")
            )
        except SchemaConversionException as e:
            self.log.exception(e)
            self.display_error_msg(self.alteryx_engine.xmsg("Schema exception: @1", e.message))
        except Exception as e:
            self.log.exception(e)
            self.display_error_msg(self.alteryx_engine.xmsg("Generic error received: @1", str(e)))
        return False

    def _push_records_from_query(self, client, update_only):
        job_config = bigquery.QueryJobConfig()
        job_config.use_legacy_sql = self.sql_type == 'LEGACY_SQL'

        # Data Query
        custom_query_to_run = self.custom_query
        if update_only and "limit" not in self.custom_query.lower():
            custom_query_to_run = self.custom_query.rstrip(';') + " limit 1;"

        query_job = client.query(custom_query_to_run, job_config=job_config)
        query_result = query_job.result()

        schema_reader = BigQuerySchemaReader(self.alteryx_engine, query_result, self.max_field_length)
        if update_only:
            return schema_reader, query_result, 1
        return schema_reader, query_result, 0

    def _push_records_from_table(self, client, update_only):
        dataset_ref = client.dataset(self.dataset_id)
        table_ref = dataset_ref.table(self.table_id)
        table = client.get_table(table_ref)
        schema_reader = BigQuerySchemaReader(self.alteryx_engine, table, self.max_field_length)
        row_iterator = None
        if not update_only:
            if self.should_limit_records == 'True' and self.record_limit > 0:
                row_iterator = client.list_rows(table, max_results=self.record_limit)
            else:
                row_iterator = client.list_rows(table)
        streaming_buffer = table.streaming_buffer
        total_rows = table.num_rows + streaming_buffer.estimated_rows if streaming_buffer else table.num_rows
        return schema_reader, row_iterator, min(total_rows, self.record_limit)
