import traceback
from typing import List

import AlteryxPythonSDK as Sdk
from connectors_core import alteryx
from connectors_core.bigquery import BigQuerySchemaWriter, SchemaConversionException
from enum import Enum
import re

from google.cloud import bigquery
from google.api_core.exceptions import NotFound, Forbidden, GoogleAPIError
from google.auth.exceptions import TransportError, RefreshError, DefaultCredentialsError

class OutputMode(Enum):
    Append = 1
    Overwrite = 2

class SchemaFieldTempl(object):
    def __init__(self, name, field_type):
        self._name = name
        self._field_type = field_type
        self._fields = []

    @property
    def name(self):
        return self._name

    @property
    def field_type(self):
        return self._field_type

    @property
    def fields(self):
        return self._fields

    def __repr__(self):
        return "SchemaFieldTempl{}".format((self._name, self._field_type, self._fields))


class BigQueryIncomingConnectionsHandler(alteryx.IncomingConnectionBase):

    def __init__(self, parent: alteryx.AyxPluginBase):
        super().__init__(parent)
        self.record_info = None  # type: Sdk.RecordInfo
        self.client = None
        self.dataset = None
        self.table = None
        self.table_ref = None
        self.schema_writer = None
        self.row_buffer = []
        self.row_insert_batch_size = None
        self.is_initialized = False
        self.output_mode = OutputMode.Append
        self.should_recreate_table = False

    def _is_valid_field_name(self, field_name):
        char_re = re.compile(r'[^a-zA-Z0-9_.]')
        field_name = char_re.search(field_name)
        return not bool(field_name)

    def _get_bigquery_type(self, ayx_field_type):
        mapping = {
            Sdk.FieldType.byte: "INTEGER",
            Sdk.FieldType.string: "STRING",
            Sdk.FieldType.wstring: "STRING",
            Sdk.FieldType.v_wstring: "STRING",
            Sdk.FieldType.v_string: "STRING",
            Sdk.FieldType.int16: "INTEGER",
            Sdk.FieldType.int32: "INTEGER",
            Sdk.FieldType.int64: "INTEGER",
            Sdk.FieldType.double: "FLOAT",
            Sdk.FieldType.float: "FLOAT",
            Sdk.FieldType.bool: "BOOLEAN",
            Sdk.FieldType.datetime: "DATETIME",
            Sdk.FieldType.date: "DATE",
            Sdk.FieldType.time: "TIME",
            Sdk.FieldType.fixeddecimal: "NUMERIC",
            Sdk.FieldType.blob: "BYTES"
        }
        try:
            return mapping[ayx_field_type]
        except KeyError as e:
            return "STRING"

    def _construct_bigquery_record_fields(self, templ_schema_fields):
        fields = []
        for i in range(len(templ_schema_fields)):
            if templ_schema_fields[i].field_type == "RECORD":
                fields.append(bigquery.SchemaField(
                    templ_schema_fields[i].name,
                    "RECORD",
                    fields = self._construct_bigquery_record_fields(templ_schema_fields[i].fields)
                ))
            else:
                fields.append(templ_schema_fields[i])
        return fields

    def _apply_ayx_field_to_schema(self, field_type, name_components, templ_schema_fields):
        top_name = name_components[0]
        if (len(name_components) > 1):
            for i in range(len(templ_schema_fields)):
                if templ_schema_fields[i].name == top_name:
                    return self._apply_ayx_field_to_schema(field_type, name_components[1:], templ_schema_fields[i].fields)
            # There is no sub-record so we have to create it
            templ_schema_fields.append(SchemaFieldTempl(top_name, "RECORD"))
            return self._apply_ayx_field_to_schema(field_type, name_components[1:], templ_schema_fields[-1].fields)

        templ_schema_fields.append(bigquery.SchemaField(top_name, field_type))

    def _retreive_output_schema(self, ayx_record):
        # Construct schema from Alteryx record
        table_schema = []
        schema_fields_templ = None # template class in order to collect fields data about record
        for i in range(0, ayx_record.num_fields):
            field = ayx_record[i]  # type: Sdk.Field
            if not self._is_valid_field_name(field.name):
                raise SchemaConversionException(
                    'Invalid column name \'{}\'. Column names must contain letters (upper or lower case), numbers, underscores or dots only in case of structures'.format(field.name)
                )
            name_components = field.name.split('.')
            if len(name_components) == 1: # we have a simple data type i.e. not RECORD
                if schema_fields_templ != None: # check if we constructed a record template previously
                    table_schema.append(
                        bigquery.SchemaField(
                            schema_fields_templ.name, "RECORD", fields = self._construct_bigquery_record_fields(schema_fields_templ.fields)
                        )
                    )
                    schema_fields_templ = None
                table_schema.append(bigquery.SchemaField(field.name, self._get_bigquery_type(field.type)))
            else: # record data type case
                if schema_fields_templ == None:
                    schema_fields_templ = SchemaFieldTempl(name_components[0], "RECORD")
                elif schema_fields_templ.name != name_components[0]: # old record is done
                    table_schema.append(
                        bigquery.SchemaField(
                            schema_fields_templ.name, "RECORD", fields = self._construct_bigquery_record_fields(schema_fields_templ.fields)
                        )
                    )
                    schema_fields_templ = SchemaFieldTempl(name_components[0], "RECORD")
                self._apply_ayx_field_to_schema(
                    self._get_bigquery_type(field.type), name_components[1:], schema_fields_templ.fields
                )

        # maybe we constructed a record template as a last field
        if schema_fields_templ != None:
            table_schema.append(
                bigquery.SchemaField(
                    schema_fields_templ.name, "RECORD", fields = self._construct_bigquery_record_fields(schema_fields_templ.fields)
                )
            )
        return table_schema

    def _flush_row_buffer_(self):
        if len(self.row_buffer) > 0:
            _run_load_job_(self.get_client(), self.row_buffer, self.table)
            self.row_buffer = []

    def get_client(self):
        if not self.client:
            self.client = bigquery.Client(project=self.parent.project_id, credentials=self.parent.credentials)
        return self.client

    def init_dataset(self):
        if self.dataset is not None:
            return
        client = self.get_client()
        dataset_ref = client.dataset(self.parent.dataset_id)
        self.dataset = client.get_dataset(dataset_ref)

    def init_table(self):
        if self.table is not None:
            return
        if self.dataset is None:
            self.init_dataset()

        client = self.get_client()
        self.table_ref = self.dataset.table(self.parent.table_id)
        self.table = client.get_table(self.table_ref)

        if self.output_mode == OutputMode.Overwrite:
            # Handle schema changes
            table_schema = self._retreive_output_schema(self.record_info)
            self.table.schema = table_schema

            # Handle clustering fields
            old_clustering_fields = self.table.clustering_fields
            if old_clustering_fields:
                new_clustering_fields = list()
                for old_field in old_clustering_fields:
                    for field in table_schema:
                        if old_field == field.name:
                            new_clustering_fields.append(old_field)
                if new_clustering_fields:
                    self.table.clustering_fields = new_clustering_fields
                else:
                    self.table.clustering_fields = None

            # Handle range partitioning
            range_partition = self.table.range_partitioning
            remove_partitioner = False
            if range_partition:
                remove_partitioner = True
                for field in table_schema:
                    if range_partition.field == field.name:
                        if field.field_type == "INTEGER" or field.field_type == "INT64":
                            remove_partitioner = False
                        break
            if remove_partitioner:
                self.table.range_partitioning = None

            # Handle time partitioning
            time_partition = self.table.time_partitioning
            remove_partitioner = False
            if time_partition and time_partition.field and time_partition.field != "_PARTITIONTIME":
                remove_partitioner = True
                for field in table_schema:
                    if time_partition.field == field.name:
                        if field.field_type == "TIMESTAMP" or field.field_type == "DATE":
                            remove_partitioner = False
                        break
            if remove_partitioner:
                self.table.time_partition = None

            self.should_recreate_table = True

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        if not self.parent.is_initialized:
            return False
        self.record_info = record_info_in
        if self.parent.output_mode == 'Append':
            self.output_mode = OutputMode.Append
        elif self.parent.output_mode == 'Overwrite':
            self.output_mode = OutputMode.Overwrite

        try:
            self.init_dataset()
            self.init_table()
            self.schema_writer = BigQuerySchemaWriter(self.table, self.record_info)
            self.row_insert_batch_size = self.parent.row_insert_batch_size
            self.is_initialized = True
            return True
        except NotFound as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "@1 Bad Request to Google API: @2", str(e.code.value), e.message
            ))
        except Forbidden as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "@1 Invalid Permissions: @2", str(e.code.value), e.message
            ))
        except GoogleAPIError as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "GoogleAPIError: @1", str(e)
            ))
        except SchemaConversionException as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                "Generic error: @1", e.message
            ))
        except TransportError as e:
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg("TransportError received. Check your proxy credentials for accuracy.")
            )
        except RefreshError as e:
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg("RefreshError received. Check your credentials for accuracy.")
            )
        except DefaultCredentialsError as e:
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg("DefaultCredentialsError received. Check your credentials for accuracy.")
            )
        except Exception as e:
            self.parent.display_error_msg(self.parent.alteryx_engine.xmsg("Generic error received: @1", str(e)))

        return False

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        try:
            if self.should_recreate_table:
                self.should_recreate_table = False
                client = self.get_client()
                # Drop old table
                client.delete_table(self.table_ref)
                # Then create a new one from ayx record schema
                self.table = client.create_table(self.table)
            row = self.schema_writer.build_row_from_record(in_record, self.parent.alteryx_engine)
        except SchemaConversionException as e:
            self.parent.display_error_msg(e.message)
            return False
        self.row_buffer.append(row)
        if len(self.row_buffer) >= self.row_insert_batch_size:
            ret = False
            try:
                self._flush_row_buffer_()
                ret = True
            except Forbidden as e:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 Invalid Permissions: @2", str(e.code.value), e.message
                ))
            except GoogleAPIError as e:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message
                ))
            except Exception:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Unhandled exception. Traceback will follow."))
                self.parent.display_error_msg(traceback.format_exc())
            return ret
        else:
            return True

    def ii_update_progress(self, d_percent: float):
        try:
            self.parent.alteryx_engine.output_tool_progress(self.parent.n_tool_id, d_percent)
        except AttributeError:
            pass

    def ii_close(self):
        try:
            self._flush_row_buffer_()
        except Forbidden as e:
            if self.parent:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 Invalid Permissions: @2", str(e.code.value), e.message
                ))
        except GoogleAPIError as e:
            if self.parent:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "@1 GoogleAPIError @2: @3", str(e.code.value), type(e).__name__, e.message
                ))
        except AttributeError:
            pass
        except Exception:
            if self.parent.is_initialized:
                self.parent.display_error_msg(self.parent.alteryx_engine.xmsg(
                    "Unhandled exception. Traceback will follow."))
                self.parent.display_error_msg(traceback.format_exc())
        self.is_initialized = False


def _run_load_job_(client: bigquery.Client, row_buffer: List[dict], table: bigquery.Table):
    job_config = bigquery.LoadJobConfig(source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
                                        schema=table.schema)
    load_job = client.load_table_from_json(row_buffer, table, job_config=job_config)

    # retry=None, timeout=None are the default parameters to result(). These can be tuned following:
    # https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.job.LoadJob.html#google.cloud.bigquery.job.LoadJob.result
    load_job.result()
