import AlteryxPythonSDK as Sdk
import connectors_core.alteryx as alteryx

from typing import Union

from connectors_core.bigquery.base import AbstractBigQueryPlugin
from .connections import BigQueryIncomingConnectionsHandler
from connectors_core.bigquery.common import BIG_QUERY_API_BASE


class BigQueryOutputEngine(AbstractBigQueryPlugin):

    def __init__(self, n_tool_id: int, alteryx_engine: Sdk.AlteryxEngine, output_anchor_mgr: Sdk.OutputAnchorManager):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)
        self.auth_type = None   # type: str
        self.service_account_json = None # type: str
        self.project_id = None # type: str
        self.dataset_id = None # type: str
        self.table_id = None # type: str
        self.credentials = None # type: google.auth.credentials.Credentials
        self.row_insert_batch_size = None
        self.incoming_connection_added = False
        self.output_mode = None

    def pi_add_incoming_connection(self, str_type: str, str_name: str) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg(self.alteryx_engine.xmsg(
                "Only one incoming connection can be accepted by this tool"
            ))
            return None
        connection_handler = BigQueryIncomingConnectionsHandler(self)
        self.incoming_connection_added = True
        return connection_handler

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)
        self.auth_type = self.get_xml_config_value(str_xml, 'authType')
        if self.auth_type == 'EndUser':
            if not self.init_end_user_credentials(str_xml):
                return
        elif self.auth_type == 'ServicePrincipal':
            if not self.init_service_principal_credentials(str_xml):
                return
        else:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid authentication mode selected: @1', self.auth_type))
            return

        self.output_mode = self.get_xml_config_value(str_xml, 'tableWriteMode')
        selected_id = self.get_xml_config_value(str_xml, 'selectedId')
        try:
            self.project_id, remainder = selected_id.split(':')
            self.dataset_id, self.table_id = remainder.split('.')
        except (AttributeError, ValueError) as e:
            self.invalidate_config(self.alteryx_engine.xmsg('Invalid table selection: please select a table.'))
        try:
            self.row_insert_batch_size = int(self.get_xml_config_value(str_xml, 'rowInsertBatchSize'))
        except TypeError:
            self.row_insert_batch_size = 500

        metadata_dict = {
            'Host': BIG_QUERY_API_BASE,
            'ServiceName': 'BigQuery',
            'PluginName': 'BigQuery',
            'Database': self.dataset_id,
            'Table': self.table_id
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor('Output')  # type: Sdk.OutputAnchor

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg(self.alteryx_engine.xmsg('This tool does not allow outgoing connections'))
        return False

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        return False
