import json
from abc import ABC

import connectors_core.alteryx as alteryx

from google.oauth2 import credentials, service_account

from connectors_core.bigquery.common import OAUTH_AUD, BIG_QUERY_SCOPE, DRIVE_SCOPE
from connectors_core.bigquery.helpers import ayx_update_proxy_environ, ayx_restore_proxy_environ


class AbstractBigQueryPlugin(alteryx.AyxPluginBase, ABC):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._previous_environ = None

    def init_end_user_credentials(self, str_xml: str):
        access_token_str = self.get_encrypted_xml_config_value(str_xml, 'accessToken')
        try:
            access_token = json.loads(access_token_str) if access_token_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Access token is not valid JSON"))
            return False
        if not access_token:
            self.invalidate_config(self.alteryx_engine.xmsg("No access token provided for end user authentication"))
            return False

        client_id = self.get_xml_config_value(str_xml, 'clientId')
        client_secret = self.get_encrypted_xml_config_value(str_xml, 'clientSecret')
        if not client_id:
            self.invalidate_config(self.alteryx_engine.xmsg('Please specify a Client ID for end user authentication'))
            return False
        if not client_secret:
            self.invalidate_config(
                self.alteryx_engine.xmsg('Please specify a Client secret for end user authentication')
            )
            return False

        try:
            self.credentials = credentials.Credentials(
                access_token['access_token'],
                refresh_token=access_token['refresh_token'],
                token_uri=OAUTH_AUD,
                client_id=client_id,
                client_secret=client_secret
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg('Invalid access token or client credentials given for end user authentication')
            )
            return False
        except Exception as e:
            self.invalidate_config(self.alteryx_engine.xmsg('Error authenticating to Google BigQuery'))
            return False
        return True

    def init_service_principal_credentials(self, str_xml: str):
        json_str = self.get_encrypted_xml_config_value(str_xml, 'serviceCredsJSONStr')
        try:
            self.service_account_json = json.loads(json_str) if json_str else None
        except json.decoder.JSONDecodeError as e:
            self.invalidate_config(self.alteryx_engine.xmsg(
                "Invalid service account keyfile. Check that you specified a valid JSON keyfile."
            ))
            return False
        if not self.service_account_json:
            self.invalidate_config(self.alteryx_engine.xmsg(
                'Service account JSON not provided. Please specify a path to your service account file.'
            ))
            return False
        try:
            scopes = [BIG_QUERY_SCOPE]
            if self.get_xml_config_value(str_xml, 'driveCheckbox') == 'True':
                scopes.append(DRIVE_SCOPE)
            self.credentials = service_account.Credentials.from_service_account_info(
                self.service_account_json,
                scopes=scopes
            )
        except ValueError as e:
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid service account keyfile. @1", str(e)))
            return False

        return True

    def pi_init(self, *args, **kwargs):
        super().pi_init(*args, **kwargs)
        self._update_proxy_environment_()

    def pi_close(self, *args, **kwargs):
        super().pi_close(*args, **kwargs)
        self._restore_proxy_environment_()
        if self.output_anchor:
            self.output_anchor.assert_close()

    def _update_proxy_environment_(self):
        self._previous_environ = ayx_update_proxy_environ(BIG_QUERY_SCOPE)

    def _restore_proxy_environment_(self):
        ayx_restore_proxy_environ(self._previous_environ)
        self._previous_environ = None
