from typing import BinaryIO, Union, Callable

import AlteryxPythonSDK as Sdk

from connectors_core.alteryx.helpers import build_record_info
from .interfaces import FileInputReader
from openpyxl import load_workbook


class ExcelReader(FileInputReader):
    def __init__(
        self,
        stream: BinaryIO,
        output_anchor: Sdk.OutputAnchor,
        alteryx_engine: Sdk.AlteryxEngine,
        stream_length_bytes: Union[int, None] = None,
        field_source_resolver: Callable[[Sdk.Field], str] = lambda f: "",
        has_header_row: bool = True,
        start_data_at_line=1,
        sheet_name: Union[str, None] = None,
    ):
        super().__init__(
            stream,
            output_anchor,
            alteryx_engine,
            stream_length_bytes,
            field_source_resolver,
        )
        self.has_header_row = has_header_row
        self.start_data_at_line = start_data_at_line
        self.sheet_name = sheet_name
        self.first_row = None

    def init(self) -> Sdk.RecordInfo:
        self.workbook = load_workbook(
            filename=self.stream, read_only=True, data_only=True
        )
        self.sheet_name = self.sheet_name or self.workbook.sheetnames[0]
        self.worksheet = self.workbook[self.sheet_name]

        try:
            row = [r.value for r in next(self.worksheet.rows)]
        except StopIteration:
            return None
        if self.has_header_row:
            self.record_info = build_record_info(
                self.alteryx_engine,
                [
                    {
                        "name": str(r),
                        "type": Sdk.FieldType.v_wstring,
                        "size": 255,  # JD-TODO
                        "scale": 0,
                        "description": None,
                    }
                    for r in row
                ],
                self.field_source_resolver,
            )
        else:
            self.record_info = build_record_info(
                self.alteryx_engine,
                [
                    {
                        "name": f"Field_{i}",
                        "type": Sdk.FieldType.v_wstring,
                        "size": 255,  # JD-TODO
                        "scale": 0,
                        "description": None,
                    }
                    for i, _ in enumerate(row)
                ],
                self.field_source_resolver,
            )
            self.first_row = row

        self.output_anchor.init(self.record_info)
        self.record_creator = self.record_info.construct_record_creator()
        self.is_initialized = True
        return self.record_info

    def read(self, n_record_limit=-1):
        record_count = 0
        if self.first_row:
            self.push_record(self.first_row)
            record_count += 1
        for record in self.worksheet.iter_rows(
            min_row=self.start_data_at_line + 1,
            max_row=None
            if n_record_limit == -1
            else self.start_data_at_line + n_record_limit,
            values_only=True,
        ):
            # replace empty values with empty string, as it would be later on converted to string "None"
            updated_record = [value if value is not None else "" for value in record]
            # if there are empty cells at the end of the row, iterator will skip those values, so lets take number of columns and fill it with empty strings
            # that way it won't happend that some nulls stay nulls while other become empty strings
            if self.record_info.num_fields > len(record):
                updated_record += ["" for i in range(self.record_info.num_fields - len(record))]
            self.push_record(updated_record)
            record_count += 1

        self.workbook.close()
        return record_count

    def push_record(self, record):
        record_info_out = self.record_info.clone()
        self.record_creator.reset()
        for field_index, field in enumerate(record):
            record_info_out[field_index].set_from_string(
                self.record_creator, str(field)
            )
        finalized_record_out = self.record_creator.finalize_record()
        self.output_anchor.push_record(finalized_record_out, no_auto_close=False)
        self.report_input_stream_progress()
